"""

``hippynn`` currently has interfaces to the following other codes:

1. ``ase`` for atomic simulation
2. ``pyseqm`` for combined ML-seqm models
3. ``schnetpack`` for usage of schnet models in ``hippynn``.

This subpackage is not available by default; you must import it explicitly.

"""
